# -----------------------------------------------------------------
# Put up a simple onscreen menu.
#
# Yes, I could have used other code for this, but I wanted to try
# rolling my own to see how hard it was.
#
# Copyright 2005-2008 Ron Dippold
# -----------------------------------------------------------------


import pygame
from pygame.locals import *
from TextSprite import TextSprite

# -----------------------------------------------------------------
# Class Menu
#
# A menu list is a [] of item dicts, in the order they will
#   be shown onscreen. Each item has:
#
# "name":  The name key - used for getting value later. This
#          is never shown to the user.
# "type":  Either 'choice' or 'action'.
# "label": What will be shown as the label for the item.
# "keys":  (tuple) of keys that will activate this item.
#          Can't be K_RETURN, K_UP, K_DOWN, K_LEFT, K_RIGHT.
#
# For "type": "choice" items -
#   "choices": A [list] of choices to be shown on screen
#   "choice":  Current choice - this is a value in the list,
#              not the index
#   "action": Optional callback after the choice is changed.
#             same as for "type": "action" - see below.
#
# For "type": "action" items
#   "action":   a callback function which looks like
#                 callback_func( _menu, _name )
#               _menu is the entire Menu() object,
#				_name is the "name" of the selected item
# -----------------------------------------------------------------

class Menu:

	def __init__( self, _items, _rect, _spriteset, _startitem=None,
				  _textcolor = (255,255,255), _barcolor = (255,0,0),
				  _commentcolor = (255,255,0)
				  ):
		"""
		Create a new Menu() object (but don't actually show it yet).
		_items is a list of dicts for each item to show (see docs)
		_rect is the rectangle on the screen we live in
		_spriteset is the pygame.sprite.RenderUpdates() we can add to
		_startitem is which item to start the cursor on when activated
		"""

		# save the passed in stuff
		self.items = _items
		self.area = _rect
		self.spriteset = _spriteset

		# Make fonts
		self.font = pygame.font.SysFont( None, 24 )
		self.commentfont = pygame.font.SysFont( None, 20 )

		# save colors
		self.textcolor = _textcolor
		self.barcolor = _barcolor
		self.commentcolor = _commentcolor

		# Translate the _startitem into an _items index. If we don't
		# find it, that's okay, just start at top.
		self.startitem = _startitem
		self.item = 0

		# For each item, set the "choiceidx"
		for item in self.items:
			if item["type"]!="choice":
				continue
			try:
				item["choiceidx"] = item["choices"].index(item["choice"])
			except ValueError:
				item["choiceidx"] = 0

		# for each item, make sure the keys is a tuple
		for item in self.items:
			keys = item.get("keys")
			if keys and not isinstance( keys, (list,tuple) ):
				item["keys"] = ( keys, )

		# We're still deactivated - no image
		self.active = False
		self.sprite = None
		self.comment = None


	# -----------------------------
	# Get the current choice of an item
	# -----------------------------

	def get_choice( self, _name ):
		"""
		Get the 'choice' value of the item with name _name
		"""
		for item in self.items:
			if item["name"]==_name:
				return item["choice"]
		return None

	# -----------------------------
	# Activate/deactivate
	# -----------------------------

	def is_active( self ):
		"""Return True if the menu is active (up on screen)"""
		return self.active


	def sprite_create( self ):
		# Figure out the max size/width
		maxheight, maxwidth = 0, 0
		for item in self.items:
			width, height = self.get_item_size( item )
			maxwidth = max( maxwidth, width )
			maxheight = max( maxheight, height )

		# Distance between items will be max height of any item
		self.itemheight = maxheight
		self.itemheight2 = maxheight/2

		# Width of the image is the max plus a little padding
		self.width = maxwidth+40

		# Then create a blank menu big enough to hold all items
		self.image0 = pygame.Surface( (self.width, self.itemheight * len(self.items)) )
		self.sprite = pygame.sprite.Sprite()

		self.sprite.image = self.image0.convert()
		self.sprite.rect = self.sprite.image.get_rect()
		self.sprite.rect.centerx = self.area.centerx

		# Make a place and sprite for our comment
		self.comment0 = pygame.Surface( ( self.area.width-40, self.itemheight ) )
		self.comment = pygame.sprite.Sprite()

		self.comment.image = self.comment0.convert()
		self.comment.rect = self.comment.image.get_rect()
		self.comment.rect.centerx = self.area.centerx

	def activate( self ):
		"""
		Turn the menu on - put it up on screen.
		"""

		# If we're already active, just bail. Otherwise set us active.
		if self.active:
			return
		self.active = True

		# Make a sprite if we need to
		if not self.sprite:
			self.sprite_create()

		# Set the cursor location - this also draws the menu
		self.set_cursor( self.startitem, _forcedraw=True )

		# Add it to the spriteset
		self.spriteset.add( self.sprite )
		self.spriteset.add( self.comment )
			
	def deactivate( self ):
		"""
		Remove the menu from the screen.
		"""
		# nothing to do if we're not active
		if not self.active:
			return
		self.active = False

		# Just removing from the sprite set makes it go away
		self.sprite.kill()
		self.comment.kill()

		# kill off the memory for the image and sprite
		del self.comment
		del self.sprite
		del self.image0
		self.sprite = None
		self.comment = None

	# -----------------------------
	# Do the real work -
	#    draw the menu
	# -----------------------------

	@staticmethod
	def get_item_text( _item, _choice=None ):
		"""
		Generate the text for an item based on its type and values.
		If _choice is given it will be used for the choice value, otherwise
		   that will be taken from the current item value. This is used
		   mostly for the get_item_size
		"""
		if _item["type"]=="choice":
			if not _choice:
				_choice = _item["choice"]
			text = _item["label"] + " : " + _choice
		else:
			text = _item["label"]
		return text

	def get_item_size( self, _item ):
		"""
		Return the maximum (x,y) size for the given _item, which means
		trying all the different choices if necessary to find the largest.
		"""

		# If it's a choice we have to try all of them
		if _item["type"]=="choice":
			size = (0,0)
			for choice in _item["choices"]:
				size0 = self.font.size( Menu.get_item_text( _item, choice ) )
				size = ( max( size[0], size0[0] ), max( size[1], size0[1] ) )
		else:
			# Otherwise there's just one thing to try
			size = self.font.size( Menu.get_item_text( _item ) )

		return size

	def draw_item( self, _image, _item, _pos ):
		"""
		Draw a single line item on the menu
		"""

		# if this is the cursor item, fill the background
		if _item == self.items[self.item]:
			_image.fill( self.barcolor, ( 0, _pos[1], self.width, self.itemheight ) )

		# Render the text and blit it to the menu, centered
		textimg = self.font.render( Menu.get_item_text( _item ), True, self.textcolor )
		x = _pos[0] - (textimg.get_width()/2)
		_image.blit( textimg, ( x, _pos[1] ) )

	
	def draw_menu( self ):
		"""
		redraw the entire menu
		"""
		# blank out current image
		self.image0.fill( (0,0,0 ) )

		# center top of first line of text
		x = self.image0.get_width()/2
		y = 0

		# Write each text on the screen
		for item in self.items:
			self.draw_item( self.image0, item, (x,y) )
			y += self.itemheight

		self.sprite.image = self.image0.convert()
		self.sprite.image.set_colorkey( (0,0,0), RLEACCEL )
		self.sprite.rect.centery = self.area.centery - 10

		# Now do the comment line

		# blank it and then write the text if there is any
		self.comment0.fill( (0,0,0 ) )
		item = self.items[self.item]
		comment = item.get( "comments" )
		if comment:
			comment = comment[ item.get("choiceidx",0) ]
		if comment:
			textimg = self.commentfont.render( comment, True, self.commentcolor )
			x = ( self.comment0.get_width() - textimg.get_width() ) /2
			self.comment0.blit( textimg, ( x, 0 ) )

		# convert the temporary surface to the actual image, make it transparent
		self.comment.image = self.comment0.convert()
		self.comment.image.set_colorkey( (0,0,0), RLEACCEL )
		self.comment.rect.top = self.sprite.rect.bottom + 10
		
	# -----------------------------
	# Move cursor around
	# -----------------------------

	def move_cursor( self, _dir ):
		"""
		Move the cursor.
		_dir is either -1, 0, or 1 for direction of movement.
		Redraws the menu.
		"""

		# Figure out which new item we're dealing with and let
		# set_cursor do the real work.
		self.set_cursor( ( self.item + _dir ) % len(self.items) )

	def set_cursor( self, _item, _forcedraw=False ):
		"""
		If _item is None then cursor is at the top.
		if _item is an int then set the cursor to an absolute index.
		If it's an _item, then set it to the index of that item.
		Redraws the menu if it is a different item.
		"""

		current = self.item

		# If _item is None, then just go to the top
		if _item == None:
			self.item = 0

		# If it's an int, then it's just an absolute index
		elif isinstance( _item, int ):
			self.item = _item

		# if it's a name, then look for it
		elif isinstance( _item, str ):
			self.item = 0
			for i in range( len( self.items ) ):
				if self.items[i]["name"] == _item:
					self.item = i
					break
	
		# Otherwise look for a matching object
		else:
			self.item = 0
			for i in range( len( self.items ) ):
				if self.items[i] == _item:
					self.item = i
					break

		# change the cursor if it hasn't changed
		if _forcedraw or (current != self.item):
			self.draw_menu()
		

	def change_choice( self, _dir ):
		"""
		Move the choice 'left' (_dir==-1) or 'right'(_dir==1)'
		in the choices list.
		"""

		# Get where the cursor is
		item = self.items[self.item]

		# Only if it's really a choice type, change the value
		if item["type"]=="choice":
			# change the value
			item["choiceidx"] = ( item["choiceidx"]+_dir ) % len( item["choices"] )
			item["choice"] = item["choices"][item["choiceidx"]]

			# redraw the menu
			self.draw_menu()

			# Do the action, if there is one
			if item.get("action"):
				item["action"]( self, item )

	# -----------------------------
	# Event handling - the meat
	# -----------------------------

	def do_item( self, _item ):
		self.set_cursor( _item )
		if _item["type"]=="action":
			_item["action"]( self, _item )
		elif _item["type"]=="choice":
			self.change_choice( 1 )

	def event( self, _event ):
		"""
		Handle game events. Just key down for now.
		"""
		if _event.type != KEYDOWN:
			return

		if _event.key == K_UP:
			self.move_cursor( -1 )
		elif _event.key == K_DOWN:
			self.move_cursor( 1 )

		elif _event.key == K_LEFT:
			self.change_choice( -1 )
			pass
		elif _event.key == K_RIGHT:
			self.change_choice( 1 )
			pass

		elif _event.key == K_RETURN:
			self.do_item( self.items[self.item] )
			

		# look for the key in any of the item activation lists
		else:
			for item in self.items:
				keys = item.get("keys")
				if keys and _event.key in keys:
					self.do_item( item )
