# -----------------------------------------------------------------------------
# A 'Missile' in this case is anything that goes in one direction till it hits
# the edge of the play area, or we kill it externally because it collided
# with something.
#
# Copyright 2005 Ron Dippold
# -----------------------------------------------------------------------------


import pygame.sprite, pygame.image
from   pygame.locals import *
import math

from ImageSet import ImageSet
import Util



# -----------------------------------------------------------------------------
#
# class BaseMissile
#
# -----------------------------------------------------------------------------


class BaseMissile( pygame.sprite.Sprite ):
	"""
	A missile is anything that's shot and then just goes on its way without
	any further notice. The only thing complicated is that it may have
	multiple frames.
	"""

	# if we go outside here, we're dead
	clip = None
	# A helper to see if we hit a wave
	WaveHelper = None
	# Base speed
	base_speed = 80

	@staticmethod
	def set_clip( _clip ):
		BaseMissile.clip = _clip

	@staticmethod
	def set_WaveHelper( _helper ):
		BaseMissile.WaveHelper = _helper

	@staticmethod
	def set_base_speed( _speed ):
		BaseMissile.base_speed = _speed

	# --------------------------
	# New Missile sprite
	# --------------------------
	def __init__( self, _imageset, _pos, _speed, _callback=None,
				  _check_waves=False, _check_clip=True ):
		"""
		New Missile type
		_imageset = ImageSet() to use as base. All same size!
		   NOTE - we do a shared_copy(), you don't have to
		_pos = top left starting position
		_speed = speed
		_callback = routine to call when we're done
		_check_waves is True if we see if we've hit the waves
		_check_clip is True if we kill ourself when we get out of clip area

		"""
		# Always init the sprite
		pygame.sprite.Sprite.__init__( self )

		self.id = self.__class__.__name__
		self.pos = list(_pos)
		self.speed = _speed
		self.callback = _callback

		self.check_waves = _check_waves
		self.hit_waves = False
		
		self.check_clip = _check_clip
		self.hit_clip = False

		self.imageset = _imageset.shared_copy()
		self.image, self.rect = self.imageset.newframe( 0, self.pos )
		
	#
	# Move it depending on how much time has passed
	#
	def update( self, seconds ):
		self.pos[0] += self.speed[0] * seconds
		self.pos[1] += self.speed[1] * seconds

		self.image, self.rect = self.imageset.newframe( seconds, self.pos )

		# If we're no longer in the play area, we're done
		if self.check_clip and not self.rect.colliderect( BaseMissile.clip ):
			self.hit_clip = True
			self.kill()

		# if we're supposed to check for it, see if we hit a wave
		if self.check_waves and self.pos[1]>=self.WaveHelper.min_top():
			wavey = self.WaveHelper.top( self.rect.centerx )
			if self.rect.bottom >= wavey:
				self.hit_waves = True
				self.pos[1] = wavey
				self.kill()

	# Our overridden kill - to do the callback
	def kill( self ):
		# remove ourself from any active groups
		pygame.sprite.Sprite.kill( self )
		# If we have a callback, use that
		if self.callback:
			self.callback( self )
			
# -----------------------------------------------------------------------------
#
# class Sam - our surface to air missile (and its flame)
#
# -----------------------------------------------------------------------------


class Sam( BaseMissile ):
	"""
	The Surface to Air missile we shoot upwards
	"""

	imageset = None
	sound = None

	@staticmethod
	def preload():
		Sam.imageset = Util.imageset_loader( 'sam_u' )
		Sam.ysize = Sam.imageset.get_height()
		Sam.sound = Util.sound_loader( 'sam' )

	def __init__( self, _pos, _callback=None ):
		"""
		New Surface to Air Missile
		_pos = (x,y) of top center
		"""

		if not Sam.imageset:
			Sam.preload()

		BaseMissile.__init__( self,
							  _imageset = Sam.imageset,
							  _pos = _pos,
							  _speed = 	( 0, -BaseMissile.clip.height ),
							  _callback=_callback )
		Sam.sound.play()


class Sam_f( BaseMissile ):
	"""
	Flame for our surface to air missile
	"""

	imageset = None
	
	@staticmethod
	def preload():
		Sam_f.imageset = Util.imageset_loader( "sam_f_u" )

	def __init__( self, _sam ):
		"""
		New Surface to Air Missile
		_sam = Sam() we're attached to
		"""

		# load the image and get the x center position
		if not Sam_f.imageset:
			Sam_f.preload()

		BaseMissile.__init__( self,
							  _imageset = Sam_f.imageset,
							  _pos =	[ _sam.pos[0], _sam.pos[1] + _sam.ysize ],
							  _speed = 	_sam.speed,
							  _check_clip = False
							  )


# -----------------------------------------------------------------------------
#
# class Bullet - Copter bullet
#
# -----------------------------------------------------------------------------

class Bullet( BaseMissile ):
	"""
	Copter bullet
	"""

	base_speed = 150
	imageset = None

	@staticmethod
	def preload():
		Bullet.imageset = Util.imageset_loader( "bullet" )
		Bullet.sound = Util.sound_loader( "bullet" )

	def __init__( self, _pos, _speed, _callback=None, _sound=True ):
		"""
		New helicopter bullet
		_pos = (x,y) of center
		_speed = speed normalized to 0-1 in each direction
		"""

		if not Bullet.imageset:
			Bullet.preload()

		BaseMissile.__init__( self,
							  _imageset = Bullet.imageset,
							  _pos =  _pos,
							  _speed = 	[_speed[0]*Bullet.base_speed, _speed[1]*Bullet.base_speed ],
							  _callback = _callback,
							  _check_waves = True
							  )
		Bullet.sound.play()


# -----------------------------------------------------------------------------
#
# class Bomb - dropped plane bomb
#
# -----------------------------------------------------------------------------

class Bomb( BaseMissile ):
	"""
	dropped bomb
	"""

	base_speed = 60
	imageset = None
	sound_list = {}

	@staticmethod
	def preload():
		Bomb.imageset = Util.imageset_loader( "bomb_d" )
		Bomb.sound = Util.sound_loader( "bomb" )

	def sound_play( self ):
		"""Only play up to 4 bomb sounds at once"""
		if len( Bomb.sound_list )<4:
			Bomb.sound.play()
		Bomb.sound_list[self] = True

	def kill( self ):
		# Remove us from playing list
		if Bomb.sound_list.has_key( self ):
			del Bomb.sound_list[self]
		BaseMissile.kill( self )

	def __init__( self, _pos, _callback=None ):
		"""
		New dropped bomb
		_pos = (x,y) of center
		"""

		# load the image and get the x center position
		if not Bomb.imageset:
			Bomb.preload()

		BaseMissile.__init__( self,
							  _imageset = Bomb.imageset,
							  _pos =	( _pos[0], _pos[1]+Bomb.imageset.get_height()-2 ),
							  _speed = 	[0, Bomb.base_speed ],
							  _callback = _callback,
							  _check_waves = True,
							  )
		Bomb.sound_play(self)


# -----------------------------------------------------------------------------
#
# class Cruise - a cruise missile
#
# -----------------------------------------------------------------------------


class Cruise( BaseMissile ):
	"""
	Sideways fired cruise missile
	"""

	imagesets = None
	
	@staticmethod
	def preload():
		Cruise.imagesets = { }
		imageset = Util.imageset_loader( "cruise_r" )
		Cruise.imagesets[1] = imageset
		Cruise.imagesets[-1] = imageset.flip( True, False )
		Cruise.xsize = imageset.get_width()
		Cruise.sound = Util.sound_loader( "cruise" )

	def __init__( self, _firstpass, _dir, _callback=None ):
		"""
		_firstpass = If True, this is first (top) pass
		_dir = -1 or 1
		_callback = callback when dead
		"""

		# load the image(s) and get the y center position
		if not Cruise.imagesets:
			Cruise.preload()

		imageset = Cruise.imagesets[_dir]

		#if true we're only doing the top pass. If false, the bottom pass
		self.first_pass = _firstpass
		if _firstpass:
			ypos = 20
		else:
			ypos = BaseMissile.WaveHelper.min_top()-imageset.get_height()/2-2


		# based on direction, get starting position and image
		self.dir = _dir
		if _dir<0:
			pos = ( BaseMissile.clip.right, ypos )
		else:
			pos = ( 0, ypos )

		BaseMissile.__init__( self,
							  _imageset = imageset,
							  _pos = pos,
							  _speed = 	( _dir * BaseMissile.base_speed*1.5, 0 ),
							  _callback=_callback )
		if self.first_pass:
			Cruise.sound.play(1)


	def score( self ):
		"""We're worth this much dead"""
		return 100


class Cruise_f( BaseMissile ):
	"""
	Flame for our cruise missile
	"""

	imagesets = None

	@staticmethod
	def preload():
		Cruise_f.imagesets = {}
		imageset = Util.imageset_loader( "cruise_f_r" ) 
		Cruise_f.imagesets[1] = imageset
		Cruise_f.imagesets[-1] = imageset.flip( True, False )
		Cruise_f.xsize = imageset.get_width()

	def __init__( self, _cruise ):
		"""
		_cruise = Cruise() we're attached to
		"""

		# load the image and get the x center position
		if not Cruise_f.imagesets:
			Cruise_f.preload()

		# Our position is relative to the cruise missile
		if _cruise.dir<0:
			pos = [ _cruise.pos[0] + _cruise.xsize, _cruise.pos[1] ]
		else:
			pos = [ _cruise.pos[0] - _cruise.xsize, _cruise.pos[1] ]

		BaseMissile.__init__( self,
							  _imageset = Cruise_f.imagesets[ _cruise.dir ],
							  _pos =	pos,
							  _speed = 	_cruise.speed,
							  _check_clip = False,
							  )




# -----------------------------------------------------------------------------
#
# class Mine - we completely override the update() method
#
# -----------------------------------------------------------------------------
class Mine( BaseMissile ):
	"""
	Just a floating mine that drifts slowly along, from one side of the
	screen, out the other, and back.

	"""

	# Will hold the images
	imageset = None

	# --------------------------
	# New Sprite
	# --------------------------
	def __init__( self, _pos, _wavespeed, _clip ):
		"""
		New Mine
		_pos = starting position
		_wavespeed = speed of waves, so we can drift along
		_clip = clip area - when we go out one side we return on the other
		"""
		# load the images if needed
		if not Mine.imageset:
			Mine.imageset = Util.imageset_loader( ("mine") )

		BaseMissile.__init__( self,
							  _imageset =   Mine.imageset,
							  _pos =		_pos,
							  _speed = 		( -_wavespeed/4, 0 ),
							  )
							  
	#
	# Move it depending on how much time has passed
	#
	def update( self, seconds ):
		self.pos[0] += self.speed[0] * seconds

		# Make it bob up and down a bit, slowly
		ypos = self.pos[1] + 2 * math.cos(self.pos[0]/10)
		
		self.image, self.rect = self.imageset.newframe( seconds, ( self.pos[0], ypos ) )

		# wrap from one side to the other
		if not self.rect.colliderect( self.clip ):
			if self.speed[0]>0:
				self.pos[0] = 1-self.imageset.get_width()/2
			else:
				self.pos[0] = self.clip.right+self.imageset.get_width()/2-1


# -----------------------------------------------------------------------------
#
# class Bomber - a 'plane', but really a missile
#
# -----------------------------------------------------------------------------


class Bomber( BaseMissile ):
	"""
	bomber that flies from side to side, drops bombs
	"""

	imagesets = None

	@staticmethod
	def preload():
		Bomber.imagesets = {}
		imageset = Util.imageset_loader( "bomber_r" )
		Bomber.imagesets[1] = imageset
		Bomber.imagesets[-1] = imageset.flip( True, False )
		Bomber.sound = Util.sound_loader( "bomber" )

	def __init__( self, _ypos, _dir, _bomb_distance, _bomb_callback, _dead_callback=None ):
		"""
		_ypos = y position
		_dir = -1 or 1
		_bomb_distance = how often to drop a bomb (in pixels)
		_bomb_callback = callback when we want to drop a bomb
		_dead_callback = callback when dead
		"""

		# load the image(s) and get the y center position
		if not Bomber.imagesets:
			Bomber.preload()

		# based on direction, get starting position and image
		self.dir = _dir
		if _dir<0:
			pos = ( BaseMissile.clip.right, _ypos )
		else:
			pos = ( 0, _ypos )

		BaseMissile.__init__( self,
							  _imageset = Bomber.imagesets[_dir],
							  _pos = pos,
							  _speed = 	( _dir * BaseMissile.base_speed*1.5, 0 ),
							  _callback=_dead_callback )

		self.bomb_callback = _bomb_callback
		self.bomb_distance = _bomb_distance
		self.bomb_last = pos[0]

		Bomber.sound.play(-1)

	def kill( self ):
		Bomber.sound.fadeout(250)
		BaseMissile.kill( self )

	def update( self, seconds ):

		# Drop a bomb every so often
		if abs(self.bomb_last-self.pos[0]) >= self.bomb_distance:
			self.bomb_last = self.pos[0]
			self.bomb_callback( self, self.rect.midbottom )

		BaseMissile.update( self, seconds )

	def score( self ):
		"""We're worth this much dead"""
		return 100

	


