# -----------------------------------------------------------------------------
#
# This holds the Ship and ShipCount classes.
#
# Copyright 2005 Ron Dippold
# -----------------------------------------------------------------------------
import pygame.sprite, pygame.image
from pygame.locals import *

import Game
import Util

# -----------------------------------------------------------------------------
#
# class Ship
#
# -----------------------------------------------------------------------------


class Ship( pygame.sprite.Sprite ):
	"""
	Our ship
	"""

	# ----------------------------
	# Class 'static' variables - load the image files up front
	# ----------------------------

	# Ship has 5 orientations - 1/2 left, 1/4 left, flat, 1/4 right, 1/2 right
	imagesets = ()

	# Our horizontal speed
	speed = 80

	# different states we can be in
	( STATE_OK, STATE_DYING, STATE_DEAD ) = range(3)

	#
	# Figure out the min and max x positions the ship can have
	#
	@staticmethod
	def __init_minmax( _xmax ):
		"""After loading all the images, figure out our x boundaries"""
		Ship.xmin = 0
		Ship.xmax = _xmax
		for i in Ship.imagesets:
			cx = i.get_center()[0]
			Ship.xmin = max( Ship.xmin, cx)
			Ship.xmax = min( Ship.xmax, _xmax-(i.get_width()-cx) )


	@staticmethod
	def preload( _xmax ):
		c = Util.imageset_loader( "ship_c" )
		r = Util.imageset_loader( "ship_r" )
		rr = Util.imageset_loader( "ship_rr" )

		# l and ll are just x flipped r and rr
		l = r.flip( True, False )
		ll = rr.flip( True, False )

		# Make a list of all the imagesets
		Ship.imagesets = ( ll, l, c, r, rr )

		# Figure out our boundaries
		Ship.__init_minmax( _xmax )

		Ship.sink_sound = Util.sound_loader( "ship_sink" )

	@staticmethod
	def unload():
		Ship.imagesets = ()
		

	# --------------------------
	# New Ship sprite
	# --------------------------
	def __init__( self, _base_speed, _xpos, _xmax, _WaveHelper, _keys ):
		"""
		Create a new Ship (there should be only one!)
		_base_speed is our ship's maximum (and only!) speed
		_xpos is only the center x position, y is implied from waves
		_xmax is the (raw) right edge of the field - we'll compensate for size
		_WaveHelper is reference to top/Slope calculating object
		_keys is [ leftkey, rightkey, firekey ]
		"""
		# Always init the sprite
		pygame.sprite.Sprite.__init__( self )

		# load our images if we haven't
		if not Ship.imagesets:
			Ship.preload( _xmax )

		# Save the reference to the helper object
		self.WaveHelper = _WaveHelper

		# Save the keys we'll use
		self.keys = _keys

		# Create a little blank image we can use to flash invulnerable
		self.flash = pygame.Surface( (1,1) )
		self.flash.fill( (0,0,0) )
		self.flash.set_colorkey( (0,0,0), RLEACCEL )
		self.invulnerable = 0.0

		# set our position, good state
		self.base_speed = _base_speed
		self.reset( _xpos )
		# not going anywhere yet
		self.direction_reset()
		self.update( 0 )

	def reset( self, _xpos ):
		"""
		Reset the ship to a healthy state and a known x position
		"""
		# new x position
		self.xpos = _xpos

		# We're not dying yet
		self.state = Ship.STATE_OK

	def get_xpos( self ):
		return int( self.xpos )
	
	def is_invulnerable( self ):
		return self.invulnerable > 0.0

	def set_invulnerable( self, _time ):
		self.invulnerable = _time
	
	def is_dying( self ):
		return self.state == Ship.STATE_DYING

	def is_dead( self ):
		return self.state == Ship.STATE_DEAD

	def die( self, _bottom ):
		self.state = Ship.STATE_DYING
		# leap up, then fall back
		self.death_speed = [ self.dir_val*Ship.speed*.5, -160 ]
		# choose which sprite to start with, then we rotate from here
		self.death_which = 2
		self.death_time = 0.0
		self.death_interval = 0.1
		self.death_bottom = _bottom
		self.sink_sound_played = False

	def update( self, seconds ):
		"""
		Move the ship after fractional seconds
		x position of the ship is the center
		y position and image depend entirely on the	position of the waves!
		"""

		# If we're dead, all we can do is wait for UI to reset us
		if self.state == Ship.STATE_DEAD:
			return

		# We're dying! Flip around and around
		elif self.state == Ship.STATE_DYING:
			# fall
			self.xpos += self.death_speed[0] * seconds
			self.ypos += self.death_speed[1] * seconds
			# add gravity to fall
			self.death_speed[1] += 120 * seconds
			# rotate sprite every so often
			self.death_time += seconds
			if self.death_time > self.death_interval:
				self.death_time -= self.death_interval
				self.death_which = ( self.death_which + 1 ) % len(self.imagesets)
			which = self.death_which

			# Play death sound?
			if not self.sink_sound_played:
				wavey = self.WaveHelper.top( self.xpos )
				if self.ypos > wavey:
					self.sink_sound_played = True
					self.sink_sound.play()

			# If we've fallen enough, go to dead state
			if self.ypos > (self.death_bottom+2000):
				self.state = Ship.STATE_DEAD

		# normal movement
		else:
			# reduce our invulnerability
			self.invulnerable -= seconds
				
			# Move left or right (or not at all), then clamp it
			self.xpos += self.dir_val*Ship.speed * seconds
			if self.xpos < Ship.xmin:
				self.xpos = Ship.xmin
			elif self.xpos > Ship.xmax:
				self.xpos = Ship.xmax

			# Get which ship sprite we want
			self.ypos, which = self.WaveHelper.both( self.xpos )

			# slopes are -2 to 2, change it to 0-4
			which += 2


		# if we're invulnerable, show blank for every other frame to flash
		if self.invulnerable>0.0 and not (self.image is self.flash):
			self.image = self.flash
		else:
			# otherwise, pick the next frame from the right frameset
			# The +3 is to put us slightly in the water instead of just on top
			self.image, self.rect = Ship.imagesets[which].newframe( seconds, (self.xpos, self.ypos+3 ) )
		
	#
	# Handle our own key events
	#
	def key_event( self, event ):


		# keep track of key up and key down even if we're dead.

		# Key pressed - we remember whether left/right are down so that if
		# the user does left down, right down, right up, we go left, right,
		# left respectively.
		if event.type == KEYDOWN:
			# left
			if event.key == self.keys[0]:
				self.dir_left = True
				self.dir_val = -1
			# right
			elif event.key == self.keys[1]:
				self.dir_right = True
				self.dir_val = 1
			# just ask the game to fire a missile
			elif (event.key==self.keys[2]) and (self.state==Ship.STATE_OK):
				Game.sam_fire( self.rect.midtop )

		# Key released
		elif event.type == KEYUP:
			# left
			if event.key == self.keys[0]:
				self.dir_left = False
				if self.dir_right:
					self.dir_val = 1
				else:
					self.dir_val = 0

			# right
			elif event.key == self.keys[1]:
				self.dir_right = False
				if self.dir_left:
					self.dir_val = -1
				else:
					self.dir_val = 0


	#
	# Rest the direction stuff after we get focus back
	#
	def direction_reset( self ):
		self.dir_left, self.dir_right = False, False
		self.dir_val = 0


	#
	# Helper function for ShipCount - get the center image
	#
	def center_image( self ):
		return self.imagesets[2]


# -----------------------------------------------------------------------------
#
# class ShipCount
#
# -----------------------------------------------------------------------------


class ShipCount( pygame.sprite.Sprite ):
	"""
	This just shows the number of ships left
	"""

	MAX = 9

	def __init__( self, _shipsprite, _count, _rect, _docenter=False ):
		"""
		_shipsprite = our created Ship(), so we can get the image
		_count = number of ships to start with
		_rect = rectangle to fit in
		_docenter = Try to center in the x direction
		"""

		pygame.sprite.Sprite.__init__( self )

		# Store our stuff, get the center image from the _shipsprite
		self.shipimage = _shipsprite.center_image()
		self.docenter = _docenter

		# Figure out where to place the ship icons
		self.xoff = self.shipimage.get_width()+5
		self.yoff = self.shipimage.get_height()+5
		self.cols = _rect.width / self.xoff
		self.rows = _rect.height / self.yoff

		# Move the rect down so the lowest ship is at the bottom
		yshift = _rect.height - self.rows*self.yoff
		_rect.top = _rect.top + yshift
		_rect.height = _rect.height - yshift

		# move rect to the center
		if not _docenter:
			xshift = ( _rect.width - self.cols*self.xoff)/2
			_rect.left = _rect.left + xshift
			_rect.width = _rect.width - xshift

		# Set up initial blank image, then 'draw' it
		self.image = pygame.Surface( _rect.size )
		self.rect = _rect
		self.set_count( _count )

	def set_count( self, _count ):
		"""
		Set the number of remaining ships to a fixed value
		"""
		_count = min( _count, ShipCount.MAX )
		
		self.count = _count
		self.changed = True

		self.image.fill( (0,0,0) )
		for count in range( _count ):
			row = count / self.cols
			col = count % self.cols
			y = (self.rows-1-row) * self.yoff
			x = col*self.xoff
			if row==0 and self.docenter:
				x += (self.image.get_width()-(_count%self.cols)*self.xoff)/2
			self.image.blit( self.shipimage.get_first_frame(), ( x, y ) )

		self.image.set_colorkey( self.image.get_at((0,0)), RLEACCEL )

	def increment( self ):
		"""Add a ship, up to the MAX possible"""
		if self.count < ShipCount.MAX:
			self.set_count( self.count+1 )

	def decrement( self ):
		"""one fewer ships, down to 0"""
		if self.count >0:
			self.set_count( self.count-1 )

	def get_count( self ):
		"""
		Return number of ships left
		"""
		return self.count

	def has_changed( self ):
		"""
		Returns True if our self.image has changed, and resets
		to false.
		"""
		self.changed = False
		return True
	
